from unimport.meta import MakeSingletonWithParams

__all__ = (
    "UnimportBaseException",
    "UnknownConfigKeyException",
    "ConfigFileNotFound",
)


class UnimportBaseException(Exception, metaclass=MakeSingletonWithParams):
    pass


class UnknownConfigKeyException(UnimportBaseException):
    def __init__(self, key: str) -> None:
        self.key = key

    def __str__(self):
        return f"Unknown config key '{self.key}', please read the documentation for more information."


class ConfigFileNotFound(FileNotFoundError, UnimportBaseException):
    def __init__(self, config_file: str) -> None:
        self.config_file = config_file

    def __str__(self):
        return f"Config file not found '{self.config_file}'"


class UnsupportedConfigFile(UnimportBaseException):
    def __init__(self, config_file: str) -> None:
        self.config_file = config_file

    def __str__(self):
        return f"Unsupported config file '{self.config_file}'"



from __future__ import annotations

from pathlib import Path

from unimport.color import paint
from unimport.enums import Color
from unimport.statement import Import, ImportFrom

__all__ = ("check",)


def check(path: Path, unused_imports: list[Import | ImportFrom], use_color: bool) -> None:
    for imp in unused_imports:
        if isinstance(imp, ImportFrom) and imp.star and imp.suggestions:
            context = (
                paint(f"from {imp.name} import *", Color.RED, use_color)
                + " -> "
                + paint(
                    f"from {imp.name} import {', '.join(imp.suggestions)}",
                    Color.GREEN,
                    use_color,
                )
            )
        else:
            context = paint(imp.name, Color.YELLOW, use_color)
        print(
            context
            + " at "
            + paint(path.as_posix(), Color.GREEN, use_color)
            + ":"
            + paint(str(imp.lineno), Color.GREEN, use_color)
        )



from __future__ import annotations

import contextlib
import dataclasses
import typing
from pathlib import Path

from unimport import commands, utils
from unimport.analyzers import MainAnalyzer
from unimport.color import paint
from unimport.config import Config
from unimport.enums import Color
from unimport.statement import Import

if typing.TYPE_CHECKING:
    from unimport.statement import ImportFrom


__all__ = ("Main",)


@dataclasses.dataclass
class _Result:
    unused_imports: list[Import | ImportFrom] = dataclasses.field(repr=False)
    path: Path
    source: str
    encoding: str
    newline: str | None = None


@dataclasses.dataclass
class Main:
    argv: typing.Sequence[str] | None = None

    config: Config = dataclasses.field(init=False)
    is_syntax_error: bool = dataclasses.field(init=False, default=False)
    is_unused_imports: bool = dataclasses.field(init=False, default=False)
    refactor_applied: bool = dataclasses.field(init=False, default=False)

    def __post_init__(self):
        self.config = self.argv_to_config()

    def argv_to_config(self) -> Config:
        import sys

        from unimport.config import ParseConfig

        return ParseConfig.parse_args(
            commands.generate_parser().parse_args(self.argv if self.argv is not None else sys.argv[1:])
        )

    @contextlib.contextmanager
    def analysis(self, source: str, path: Path) -> typing.Iterator:
        analyzer = MainAnalyzer(
            source=source,
            path=path,
            include_star_import=self.config.include_star_import,
        )
        try:
            analyzer.traverse()
        except SyntaxError as exc:
            print(
                paint(str(exc), Color.RED, self.config.use_color)
                + " at "
                + paint(path.as_posix(), Color.GREEN, self.config.use_color)
            )
            self.is_syntax_error = True

        try:
            yield
        finally:
            analyzer.clear()

    def get_results(self) -> typing.Iterator[_Result]:
        for path in self.config.get_paths():
            source, encoding, newline = utils.read(path)

            with self.analysis(source, path):
                unused_imports = list(Import.get_unused_imports(include_star_import=self.config.include_star_import))
                if self.is_unused_imports is False:
                    self.is_unused_imports = unused_imports != []

                yield _Result(unused_imports, path, source, encoding, newline)

    def check(self, result: _Result) -> None:
        commands.check(result.path, result.unused_imports, self.config.use_color)

    def remove(self, result: _Result, refactor_result):
        commands.remove(
            result.path,
            result.encoding,
            result.newline,
            refactor_result,
            self.config.use_color,
        )
        self.refactor_applied = True

    def diff(self, result, refactor_result):
        return commands.diff(result.path, result.source, refactor_result, self.config.use_color)

    @staticmethod
    def permission(result) -> bool:
        return commands.permission(result.path, result.encoding)

    @classmethod
    def run(cls, argv: typing.Sequence[str] | None = None) -> Main:
        from unimport.refactor import refactor_string

        self = cls(argv)
        for result in self.get_results():
            if self.config.check:
                self.check(result)
            if any((self.config.diff, self.config.remove)):
                refactor_result = refactor_string(source=result.source, unused_imports=result.unused_imports)
                if self.config.diff:
                    exists_diff = self.diff(result, refactor_result)
                    if self.config.permission and exists_diff:
                        self.config.remove = self.permission(result)
                if self.config.remove and result.source != refactor_result:
                    self.remove(result, refactor_result)
        return self

    def exit_code(self):
        from unimport.utils import return_exit_code

        return return_exit_code(
            is_unused_imports=self.is_unused_imports,
            is_syntax_error=self.is_syntax_error,
            refactor_applied=self.refactor_applied,
        )



import ast
import contextlib

from unimport import constants as C
from unimport import typing as T
from unimport.analyzers.decarators import generic_visit
from unimport.analyzers.utils import first_parent_match, set_tree_parents
from unimport.statement import Name, Scope

__all__ = ("NameAnalyzer",)


class NameAnalyzer(ast.NodeVisitor):
    def visit_ClassDef(self, node) -> None:
        Scope.add_current_scope(node)

        self.generic_visit(node)

        Scope.remove_current_scope()

    def visit_FunctionDef(self, node: T.ASTFunctionT) -> None:
        Scope.add_current_scope(node)

        if node.type_comment is not None:
            self.join_visit(node.type_comment, node, mode="func_type")

        self.generic_visit(node)

        Scope.remove_current_scope()

    visit_AsyncFunctionDef = visit_FunctionDef

    @generic_visit
    def visit_Constant(self, node: ast.Constant) -> None:
        if isinstance(node.value, str):
            if (first_annassign_or_arg := first_parent_match(node, (ast.AnnAssign, ast.arg))) and isinstance(
                first_annassign_or_arg.annotation, ast.Constant
            ):
                self.join_visit(node.value, node)
            elif (
                first_func_parent := first_parent_match(node, *C.AST_FUNCTION_TUPLE)
            ) and first_func_parent.returns is node:
                self.join_visit(node.value, node)

    @generic_visit
    def visit_Name(self, node: ast.Name) -> None:
        if not isinstance(node.parent, ast.Attribute):  # type: ignore
            Name.register(lineno=node.lineno, name=node.id, node=node)

    @generic_visit
    def visit_Attribute(self, node: ast.Attribute) -> None:
        if not isinstance(node.value, ast.Call):
            names = []
            for sub_node in ast.walk(node):
                if isinstance(sub_node, ast.Attribute):
                    names.append(sub_node.attr)
                elif isinstance(sub_node, ast.Name):
                    names.append(sub_node.id)
            names.reverse()
            Name.register(lineno=node.lineno, name=".".join(names), node=node)

    @generic_visit
    def visit_Assign(self, node: ast.Assign) -> None:
        if node.type_comment is not None:
            self.join_visit(node.type_comment, node)

    @generic_visit
    def visit_arg(self, node: ast.arg) -> None:
        if node.type_comment is not None:
            self.join_visit(node.type_comment, node)

    @generic_visit
    def visit_Subscript(self, node: ast.Subscript) -> None:
        if (
            isinstance(node.value, ast.Attribute)
            and isinstance(node.value.value, ast.Name)
            and node.value.value.id == "typing"
        ) or (isinstance(node.value, ast.Name) and node.value.id in C.SUBSCRIPT_TYPE_VARIABLE):
            if C.PY39_PLUS:
                _slice = node.slice
            else:
                _slice = node.slice.value  # type: ignore

            if isinstance(_slice, ast.Tuple):  # type: ignore
                for elt in _slice.elts:  # type: ignore
                    if isinstance(elt, ast.Constant) and isinstance(elt.value, str):
                        self.join_visit(elt.value, elt)
            else:
                if isinstance(_slice, ast.Constant) and isinstance(_slice.value, str):  # type: ignore
                    self.join_visit(_slice.value, _slice)

    @generic_visit
    def visit_Call(self, node: ast.Call) -> None:
        if (
            (
                isinstance(node.func, ast.Attribute)
                and isinstance(node.func.value, ast.Name)
                and node.func.value.id == "typing"
                and node.func.attr == "cast"
            )
            or isinstance(node.func, ast.Name)
            and node.func.id == "cast"
        ):
            if isinstance(node.args[0], ast.Constant) and isinstance(node.args[0].value, str):
                self.join_visit(node.args[0].value, node.args[0])

    def join_visit(self, value: str, node: ast.AST, *, mode: str = "eval") -> None:
        """A function that parses the value, copies locations from the node and
        includes them in self.visit."""
        with contextlib.suppress(SyntaxError):
            tree = ast.parse(value, mode=mode, type_comments=True)
            set_tree_parents(tree, parent=node.parent)  # type: ignore
            for new_node in ast.walk(tree):
                ast.copy_location(new_node, node)
            self.visit(tree)

    def traverse(self, tree) -> None:
        self.visit(tree)



import ast
import re
from pathlib import Path

from unimport.analyzers.import_statement import ImportAnalyzer
from unimport.analyzers.importable import ImportableAnalyzer
from unimport.analyzers.name import NameAnalyzer
from unimport.analyzers.utils import get_defined_names, set_tree_parents
from unimport.statement import Import, ImportFrom, Name, Scope

__all__ = ("MainAnalyzer",)


class MainAnalyzer(ast.NodeVisitor):
    __slots__ = ("source", "path", "include_star_import")

    def __init__(self, *, source: str, path: Path = Path("<unknown file>"), include_star_import: bool = False):
        self.source = source
        self.path = path
        self.include_star_import = include_star_import

    def __enter__(self):
        self.traverse()
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.clear()

    def traverse(self) -> None:
        if self.skip_file():
            return None

        tree = ast.parse(self.source, type_comments=True)

        set_tree_parents(tree)  # set parents to tree

        Scope.add_global_scope(tree)  # add global scope of the top tree

        NameAnalyzer().traverse(tree)  # name analyzers

        ImportableAnalyzer().traverse(tree)  # importable analyzers for collect in __all__

        ImportAnalyzer(  # import analyzers
            source=self.source, include_star_import=self.include_star_import, defined_names=get_defined_names(tree)
        ).traverse(tree)

        Scope.remove_current_scope()  # remove global scope

    def skip_file(self) -> bool:
        SKIP_FILE_REGEX = "#.*(unimport: {0,1}skip_file)"

        return bool(re.search(SKIP_FILE_REGEX, self.source, re.IGNORECASE))

    @staticmethod
    def clear():
        Name.clear()
        Import.clear()
        ImportFrom.clear()
        Scope.clear()



import ast
import sys
import sysconfig

__all__ = (
    "BUILTIN_MODULE_NAMES",
    "EXCLUDE_REGEX_PATTERN",
    "GLOB_PATTERN",
    "INCLUDE_REGEX_PATTERN",
    "INIT_FILE_IGNORE_REGEX",
    "PY39_PLUS",
    "PY310_PLUS",
    "STDLIB_PATH",
    "SUBSCRIPT_TYPE_VARIABLE",
)

# REGEX
GLOB_PATTERN = r"**/*.py"
INCLUDE_REGEX_PATTERN = r"\.(py)$"
EXCLUDE_REGEX_PATTERN = r"^$"
INIT_FILE_IGNORE_REGEX = r"__init__\.py"

# TUPLE
DEF_TUPLE = (ast.FunctionDef, ast.AsyncFunctionDef, ast.ClassDef)
AST_FUNCTION_TUPLE = (ast.FunctionDef, ast.AsyncFunctionDef)

# CONF
PY39_PLUS = sys.version_info >= (3, 9)
PY310_PLUS = sys.version_info >= (3, 10)

SUBSCRIPT_TYPE_VARIABLE = frozenset(
    {
        "AbstractSet",
        "AsyncContextManager",
        "AsyncGenerator",
        "AsyncIterable",
        "AsyncIterator",
        "Awaitable",
        "Callable",
        "ChainMap",
        "ClassVar",
        "Collection",
        "Container",
        "ContextManager",
        "Coroutine",
        "Counter",
        "DefaultDict",
        "Deque",
        "Dict",
        "FrozenSet",
        "Generator",
        "IO",
        "ItemsView",
        "Iterable",
        "Iterator",
        "KeysView",
        "List",
        "Mapping",
        "MappingView",
        "Match",
        "MutableMapping",
        "MutableSequence",
        "MutableSet",
        "Optional",
        "Pattern",
        "Reversible",
        "Sequence",
        "Set",
        "SupportsRound",
        "Tuple",
        "Type",
        "Union",
        "ValuesView",
        # Python >= 3.7.
        "Literal",
        # Python >= 3.8.
        "OrderedDict",
    }
)

BUILTIN_MODULE_NAMES = frozenset(sys.builtin_module_names)
STDLIB_PATH = sysconfig.get_paths()["stdlib"]



import sys
from typing import Tuple

from unimport.enums import Color

__all__ = (
    "TERMINAL_SUPPORT_COLOR",
    "difference",
    "paint",
    "Color",
)

if sys.platform == "win32":  # pragma: no cover (windows)

    def _enable() -> None:
        from ctypes import POINTER, WINFUNCTYPE, WinError, windll
        from ctypes.wintypes import BOOL, DWORD, HANDLE

        STD_ERROR_HANDLE = -12
        ENABLE_VIRTUAL_TERMINAL_PROCESSING = 4

        def bool_errcheck(result, func, args):
            if not result:
                raise WinError()
            return args

        GetStdHandle = WINFUNCTYPE(HANDLE, DWORD)(
            ("GetStdHandle", windll.kernel32),
            ((1, "nStdHandle"),),
        )

        GetConsoleMode = WINFUNCTYPE(BOOL, HANDLE, POINTER(DWORD))(
            ("GetConsoleMode", windll.kernel32),
            ((1, "hConsoleHandle"), (2, "lpMode")),
        )
        GetConsoleMode.errcheck = (  # type: ignore[assignment, misc]
            bool_errcheck  # type: ignore[assignment]
        )

        SetConsoleMode = WINFUNCTYPE(BOOL, HANDLE, DWORD)(
            ("SetConsoleMode", windll.kernel32),
            ((1, "hConsoleHandle"), (1, "dwMode")),
        )
        SetConsoleMode.errcheck = (  # type: ignore[assignment, misc]
            bool_errcheck  # type: ignore[assignment]
        )

        # As of Windows 10, the Windows console supports (some) ANSI escape
        # sequences, but it needs to be enabled using `SetConsoleMode` first.
        #
        # More info on the escape sequences supported:
        # https://msdn.microsoft.com/en-us/library/windows/desktop/mt638032(v=vs.85).aspx
        stderr = GetStdHandle(STD_ERROR_HANDLE)
        flags = GetConsoleMode(stderr)
        SetConsoleMode(stderr, flags | ENABLE_VIRTUAL_TERMINAL_PROCESSING)

    try:
        _enable()
    except OSError:
        TERMINAL_SUPPORT_COLOR = False
    else:
        TERMINAL_SUPPORT_COLOR = True
else:  # pragma: win32 no cover
    TERMINAL_SUPPORT_COLOR = True


def paint(text: str, color: Color, use_color: bool = True) -> str:
    return color.value + text + Color.RESET.value if use_color else text


def difference(text: Tuple[str, ...], use_color: bool = True) -> str:  # pragma: no cover
    lines = list(text)
    for i, line in enumerate(lines):
        if line.startswith("+++") or line.startswith("---"):
            lines[i] = paint(line, Color.BOLD_WHITE, use_color)
        elif line.startswith("@@"):
            lines[i] = paint(line, Color.CYAN, use_color)
        elif line.startswith("+"):
            lines[i] = paint(line, Color.GREEN, use_color)
        elif line.startswith("-"):
            lines[i] = paint(line, Color.RED, use_color)
    return "\n".join(lines)



import os
import sys
import textwrap
from pathlib import Path

import pytest

from tests.utils import refactor, reopenable_temp_file
from unimport import utils
from unimport.utils import action_to_bool


@pytest.mark.parametrize(
    "path, count",
    [
        ("tests/commands", 7),
        ("tests/analyzer", 0),
        ("tests/config", 1),
        ("tests/config/configs", 0),
        ("tests/config/configs/no_unimport", 0),
        ("tests/refactor", 0),
        ("tests/test_config.py", 1),
    ],
)
def test_list_paths(path, count):
    path = Path(path)

    assert len(list(utils.list_paths(path))) == count


@pytest.mark.skipif(
    sys.platform == "win32",
    reason="Patspec version 0.10.1 and above are not supported on Windows",
)
def test_list_paths_with_gitignore():
    gitignore = textwrap.dedent(
        """\
        a
        b
        spam/**
        **/api/
        **/tests
        **/
        """
    )
    with reopenable_temp_file(gitignore) as gitignore_path:
        gitignore_patterns = utils.get_exclude_list_from_gitignore(gitignore_path)
        assert list(utils.list_paths(Path("."), gitignore_patterns=gitignore_patterns)) == []


def test_bad_encoding():
    # Make conflict between BOM and encoding Cookie.
    # https://docs.python.org/3/library/tokenize.html#tokenize.detect_encoding
    bad_encoding = "\ufeff\n# -*- coding: utf-32 -*-\nbad encoding"

    with reopenable_temp_file(bad_encoding) as tmp_path:
        assert refactor(tmp_path) == ""


def test_refactor_file():
    with reopenable_temp_file("import os") as tmp_path:
        assert refactor(tmp_path) == ""


def test_diff_file():
    with reopenable_temp_file("import os") as tmp_path:
        source = utils.read(tmp_path)[0]
        refactor_result = refactor(tmp_path)
        diff_file = utils.diff(
            source=source,
            refactor_result=refactor_result,
            fromfile=tmp_path,
        )
        diff = (
            f"--- {tmp_path.as_posix()}\n",
            "+++ \n",
            "@@ -1 +0,0 @@\n",
            "-import os",
        )
        assert diff == diff_file


def test_read():
    source = "b�se"

    with reopenable_temp_file(source) as tmp_path:
        assert (source, "utf-8", None) == utils.read(tmp_path)


def test_read_newline_native():
    source = "\n".join(
        [
            "import sys",
            "",
            "print(sys.executable)\n",
        ]
    )

    with reopenable_temp_file(source, newline=os.linesep) as tmp_path:
        assert (source, "utf-8", os.linesep) == utils.read(tmp_path)


def test_read_newline_nonnative(non_native_linesep):
    source = "\n".join(
        [
            "import sys",
            "",
            "print(sys.executable)\n",
        ]
    )

    with reopenable_temp_file(source, newline=non_native_linesep) as tmp_path:
        assert (source, "utf-8", non_native_linesep) == utils.read(tmp_path)


@pytest.mark.parametrize(
    "is_unused_imports, is_syntax_error, refactor_applied, expected_exit_code",
    [
        (True, True, True, 1),
        (True, True, False, 1),
        (True, False, True, 0),
        (True, False, False, 1),
        (True, False, False, 1),
        (False, True, False, 1),
        (False, False, False, 0),
    ],
)
def test_return_exit_code(
    is_unused_imports,
    is_syntax_error,
    refactor_applied,
    expected_exit_code,
):
    assert (
        utils.return_exit_code(
            is_unused_imports=is_unused_imports,
            is_syntax_error=is_syntax_error,
            refactor_applied=refactor_applied,
        )
        == expected_exit_code
    )


def test_action_to_bool():
    yes = ("y", "Y", "yes", "True", "t", "true", "True", "On", "on", "1")
    no = ("n", "no", "f", "false", "off", "0", "Off", "No", "N")

    for y in yes:
        assert action_to_bool(y) is True

    for n in no:
        assert action_to_bool(n) is False



import enum

__all__ = ("Emoji", "Color", "ColorSelect")


class Emoji(str, enum.Enum):
    STAR = "\U0001F929"
    PARTYING_FACE = "\U0001F973"


class Color(str, enum.Enum):
    RESET = "\033[0m"
    BLACK = "\033[30m"
    RED = "\033[31m"
    GREEN = "\033[32m"
    YELLOW = "\033[33m"
    BLUE = "\033[34m"
    MAGENTA = "\033[35m"
    CYAN = "\033[36m"
    WHITE = "\033[97m"
    BOLD_WHITE = "\033[1;37m"


class ColorSelect(str, enum.Enum):
    AUTO = "auto"
    ALWAYS = "always"
    NEVER = "never"



from __future__ import annotations

import difflib
import functools
import importlib.machinery
import importlib.util
import re
import tokenize
import typing
from pathlib import Path

from pathspec.patterns.gitwildmatch import GitWildMatchPattern

import unimport.constants as C

__all__ = (
    "get_dir",
    "get_source",
    "get_spec",
    "is_std",
    "action_to_bool",
    "get_exclude_list_from_gitignore",
    "read",
    "list_paths",
    "diff",
    "return_exit_code",
)


@functools.lru_cache(maxsize=128)
def get_dir(package: str) -> frozenset[str]:
    try:
        module = importlib.import_module(package)
    except (ImportError, AttributeError, TypeError, ValueError):
        return frozenset()
    return frozenset(dir(module))


@functools.lru_cache(maxsize=128)
def get_source(package: str) -> str | None:
    spec = get_spec(package)
    if spec is not None:
        assert isinstance(spec.loader, importlib.machinery.SourceFileLoader)
        assert isinstance(spec.loader.path, str)
        if spec.loader.path.endswith(".py"):
            return spec.loader.get_data(spec.loader.path).decode("utf-8")

    return None


@functools.lru_cache(maxsize=128)
def get_spec(package: str) -> importlib.machinery.ModuleSpec | None:
    try:
        return importlib.util.find_spec(package)
    except (ImportError, AttributeError, TypeError, ValueError):
        return None


@functools.lru_cache(maxsize=128)
def is_std(package: str) -> bool:
    """Returns True if package module came with from Python."""
    if package in C.BUILTIN_MODULE_NAMES:
        return True
    spec = get_spec(package)
    if spec and isinstance(spec.origin, str):
        return any(
            (
                spec.origin.startswith(C.STDLIB_PATH),
                spec.origin in ("built-in", "frozen"),
                spec.origin.endswith(".so"),
            )
        )

    return False


@functools.lru_cache(maxsize=3)
def action_to_bool(action: str) -> bool:
    """Convert a string representation of truth to true (True) or false (False).

    True values are 'y', 'yes', 't', 'true', 'on', and '1'; false values
    are 'n', 'no', 'f', 'false', 'off', and '0'.  Raises ValueError if
    'val' is anything else.
    """
    action = action.lower()
    if action in ("y", "yes", "t", "true", "on", "1"):
        return True
    elif action in ("n", "no", "f", "false", "off", "0"):
        return False
    else:
        raise ValueError(f"invalid truth value {action!r}")


def get_exclude_list_from_gitignore(path=Path(".gitignore")) -> list[GitWildMatchPattern]:
    """Converts .gitignore patterns to regex and return this excludes regex
    list."""

    if not path.is_file():
        return []

    gitignore_patterns: list[GitWildMatchPattern] = []
    source, _, _ = read(path)
    for line in source.splitlines():
        regex, include = GitWildMatchPattern.pattern_to_regex(line)
        if regex:
            pattern = GitWildMatchPattern(re.compile(regex), include)
            gitignore_patterns.append(pattern)

    return gitignore_patterns


def read(path: Path) -> tuple[str, str, str | None]:
    try:
        with tokenize.open(path) as stream:
            source = stream.read()
            encoding = stream.encoding
            newline = stream.newlines
    except (OSError, SyntaxError):
        return "", "utf-8", None

    # If mixed or unknown newlines, fall back to the platform default
    if not isinstance(newline, str):
        newline = None

    return source, encoding, newline


def list_paths(
    start: Path,
    *,
    include: str = C.INCLUDE_REGEX_PATTERN,
    exclude: str = C.EXCLUDE_REGEX_PATTERN,
    gitignore_patterns: list[GitWildMatchPattern] | None = None,
) -> typing.Iterator[Path]:
    include_regex, exclude_regex = re.compile(include), re.compile(exclude)
    file_names: typing.Iterable[Path] = start.glob(C.GLOB_PATTERN) if start.is_dir() else [start]

    if gitignore_patterns:
        for file_name in file_names:
            if include_regex.search(str(file_name)) and not exclude_regex.search(str(file_name)):
                for gitignore_pattern in gitignore_patterns:
                    match_file = gitignore_pattern.match_file(str(file_name))
                    if match_file:
                        break
                else:
                    yield file_name
    else:
        for file_name in file_names:
            if include_regex.search(str(file_name)) and not exclude_regex.search(str(file_name)):
                yield file_name


def diff(*, source: str, refactor_result: str, fromfile: Path = None) -> tuple[str, ...]:
    return tuple(
        difflib.unified_diff(
            source.splitlines(),
            refactor_result.splitlines(),
            fromfile=fromfile.as_posix() if fromfile else "",
        )
    )


@functools.lru_cache(maxsize=128)
def return_exit_code(*, is_unused_imports: bool, is_syntax_error: bool, refactor_applied: bool) -> int:
    """If this function changes, be sure to update this page
    https://unimport.hakancelik.dev/tutorial/other-useful-features/#exit-code-
    behavior."""

    assert not (
        is_unused_imports is False and refactor_applied is True
    ), "is_unused_imports False while refactor_applied cannot be True."

    if is_syntax_error:
        return 1
    elif is_unused_imports:
        if refactor_applied:
            return 0
        return 1

    return 0



from typing import Dict

__all__ = ("MakeSingletonWithParams",)


class MakeSingletonWithParams(type):
    _instances: Dict = {}

    def __call__(cls, *args, **kwargs):
        """Returns the same instance if args and kwargs are the same,
        otherwise, creates a new instance."""
        key = (cls, args, frozenset(kwargs.items()))
        if key not in cls._instances:
            cls._instances[key] = super().__call__(*args, **kwargs)
        return cls._instances[key]



from textwrap import dedent
from unittest import mock

import pytest

from tests.utils import reopenable_temp_file
from unimport.config import Config
from unimport.main import Main


def test_empty_main():
    main = Main(["--disable-auto-discovery-config"])

    assert main.argv == ["--disable-auto-discovery-config"]
    assert main.config == Config(disable_auto_discovery_config=True)
    assert main.is_syntax_error is False
    assert main.is_unused_imports is False
    assert main.refactor_applied is False


def test_main_run_under_path():
    source = dedent(
        """\
        import os

        """
    )
    with reopenable_temp_file(source) as temp_file:
        main = Main.run(["--disable-auto-discovery-config", "--include", temp_file.as_posix()])

    assert main.argv == ["--disable-auto-discovery-config", "--include", temp_file.as_posix()]
    assert main.config == Config(disable_auto_discovery_config=True, include=temp_file.as_posix())
    assert main.is_syntax_error is False
    assert main.is_unused_imports is False
    assert main.refactor_applied is False


@pytest.mark.parametrize("command_name", ["check", "diff", "permission", "remove"])
def test_main_command(command_name, monkeypatch):
    def mock_command(*args, **kwargs):
        if not hasattr(mock_command, "call_count"):
            mock_command.call_count = 0

        mock_command.call_count += 1

    monkeypatch.setattr(Main, command_name, mock_command)

    source = dedent(
        """\
        import os
        
        """
    )
    with reopenable_temp_file(source) as temp_file:
        Main.run([f"--{command_name}", temp_file.as_posix()])

    assert mock_command.call_count == 1, f"command_name='{command_name}'"


def test_exit_code():
    main = Main([])
    assert main.exit_code() == 0

    main.is_syntax_error = True
    assert main.exit_code() == 1

    # reset syntax error
    main.is_syntax_error = False

    main.is_unused_imports = True
    assert main.exit_code() == 1

    main.refactor_applied = True
    assert main.exit_code() == 0


@mock.patch("unimport.main.Main.permission")
def test_commands_in_run(mock_permission):
    source = dedent(
        """\
        import os

        """
    )

    mock_permission.return_value = False

    assert Main(["--disable-auto-discovery-config"]).config.remove is True
    assert Main(["--disable-auto-discovery-config"]).config.permission is False

    with reopenable_temp_file(source) as temp_file:
        main = Main.run(["--disable-auto-discovery-config", f"--permission", temp_file.as_posix()])

    assert main.config.remove is False
    assert main.config.permission is True



from __future__ import annotations

import ast

from unimport import constants as C
from unimport import typing as T
from unimport import utils
from unimport.analyzers.decarators import generic_visit
from unimport.analyzers.utils import first_parent_match, set_tree_parents
from unimport.statement import Name, Scope

__all__ = ("ImportableAnalyzer",)


class ImportableAnalyzer(ast.NodeVisitor):
    __slots__ = (
        "importable_nodes",
        "suggestions_nodes",
    )

    def __init__(self) -> None:
        self.importable_nodes: list[ast.Constant] = []  # nodes on the __all__ list
        self.suggestions_nodes: list[T.ASTImportableT] = []  # nodes on the CFN

    def traverse(self, tree):
        self.visit(tree)

        for node in self.importable_nodes:
            Name.register(lineno=node.lineno, name=node.value, node=node, is_all=True)

        self.clear()

    def visit_CFN(self, node: T.CFNT) -> None:
        Scope.add_current_scope(node)

        if not first_parent_match(node, C.DEF_TUPLE):
            self.suggestions_nodes.append(node)

        self.generic_visit(node)

        Scope.remove_current_scope()

    visit_ClassDef = visit_CFN
    visit_FunctionDef = visit_CFN
    visit_AsyncFunctionDef = visit_CFN

    @generic_visit
    def visit_Import(self, node: ast.Import) -> None:
        for alias in node.names:
            self.suggestions_nodes.append(alias)

    @generic_visit
    def visit_ImportFrom(self, node: ast.ImportFrom) -> None:
        if not node.names[0].name == "*":
            for alias in node.names:
                self.suggestions_nodes.append(alias)

    @generic_visit
    def visit_Assign(self, node: ast.Assign) -> None:
        if getattr(node.targets[0], "id", None) == "__all__" and isinstance(node.value, (ast.List, ast.Tuple, ast.Set)):
            for item in node.value.elts:
                if isinstance(item, ast.Constant) and isinstance(item.value, str):
                    self.importable_nodes.append(item)

        for target in node.targets:  # we only get assigned names
            if isinstance(target, (ast.Name, ast.Attribute)):
                self.suggestions_nodes.append(target)

    @generic_visit
    def visit_Expr(self, node: ast.Expr) -> None:
        if (
            isinstance(node.value, ast.Call)
            and isinstance(node.value.func, ast.Attribute)
            and isinstance(node.value.func.value, ast.Name)
            and node.value.func.value.id == "__all__"
        ):
            if node.value.func.attr == "append":
                for arg in node.value.args:
                    if isinstance(arg, ast.Constant) and isinstance(arg.value, str):
                        self.importable_nodes.append(arg)

            elif node.value.func.attr == "extend":
                for arg in node.value.args:
                    if isinstance(arg, ast.List):
                        for item in arg.elts:
                            if isinstance(item, ast.Constant) and isinstance(item.value, str):
                                self.importable_nodes.append(item)

    @classmethod
    def get_names(cls, package: str) -> frozenset[str]:
        if utils.is_std(package):
            return utils.get_dir(package)

        source = utils.get_source(package)
        if source:
            try:
                tree = ast.parse(source)
            except SyntaxError:
                return frozenset()
            else:
                visitor = cls()
                set_tree_parents(tree)
                visitor.visit(tree)
                return visitor.get_all() or visitor.get_suggestion()
        return frozenset()

    def get_all(self) -> frozenset[str]:
        names = set()
        for node in self.importable_nodes:
            names.add(node.value)
        return frozenset(names)

    def get_suggestion(self) -> frozenset[str]:
        names = set()
        for node in self.suggestions_nodes:  # type: ignore
            if isinstance(node, ast.Name):
                names.add(node.id)
            elif isinstance(node, ast.alias):
                names.add(node.asname or node.name)
            elif isinstance(node, C.DEF_TUPLE):
                names.add(node.name)
        return frozenset(names)

    def clear(self):
        self.importable_nodes.clear()
        self.suggestions_nodes.clear()



import functools
import re
from typing import cast

from unimport import typing as T

__all__ = ("generic_visit", "skip_import")


def generic_visit(func: T.FunctionT) -> T.FunctionT:
    @functools.wraps(func)
    def wrapper(self, node, *args, **kwargs):
        func(self, node, *args, **kwargs)
        self.generic_visit(node)

    return cast(T.FunctionT, wrapper)


def skip_import(func: T.FunctionT) -> T.FunctionT:
    SKIP_IMPORT_COMMENTS_REGEX = "#.*(unimport: {0,1}skip|noqa)"

    @functools.wraps(func)
    def wrapper(self, node, *args, **kwargs):
        source_segment = "\n".join(self.source.splitlines()[node.lineno - 1 : node.end_lineno])
        skip_comment = bool(re.search(SKIP_IMPORT_COMMENTS_REGEX, source_segment, re.IGNORECASE))
        if not any((skip_comment, self.any_import_error)):
            func(self, node, *args, **kwargs)

    return cast(T.FunctionT, wrapper)



import ast
import typing

import libcst as cst

__all__ = (
    "FunctionT",
    "ASTImportableT",
    "ASTFunctionT",
    "CFNT",
    "CSTImportT",
)

ASTImportableT = typing.Union[ast.AsyncFunctionDef, ast.Attribute, ast.ClassDef, ast.FunctionDef, ast.Name, ast.alias]
ASTFunctionT = typing.TypeVar("ASTFunctionT", ast.FunctionDef, ast.AsyncFunctionDef)

CFNT = typing.TypeVar("CFNT", ast.ClassDef, ast.FunctionDef, ast.AsyncFunctionDef, ast.Name)
CSTImportT = typing.TypeVar("CSTImportT", cst.Import, cst.ImportFrom)

FunctionT = typing.TypeVar("FunctionT", bound=typing.Callable[..., typing.Any])



import argparse

from unimport import __description__
from unimport.commands import options
from unimport.enums import Emoji

__all__ = ("generate_parser",)


def generate_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="unimport",
        description=__description__,
        epilog=f"Get rid of all unused imports {Emoji.PARTYING_FACE}",
    )
    exclusive_group = parser.add_mutually_exclusive_group(required=False)

    options.add_color_option(parser)
    options.add_sources_option(parser)
    options.add_check_option(parser)
    options.add_config_option(parser)
    options.add_disable_auto_discovery_config_option(parser)
    options.add_include_option(parser)
    options.add_exclude_option(parser)
    options.add_gitignore_option(parser)
    options.add_ignore_init_option(parser)
    options.add_include_star_import_option(parser)
    options.add_diff_option(parser)
    options.add_remove_option(exclusive_group)
    options.add_permission_option(exclusive_group)
    options.add_version_option(parser)

    return parser
